# core-result-checker
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import configparser
from datetime import datetime
from functools import partial
import logging
import os

import kombu
from kombu.pools import producers
from swiftclient.service import SwiftService
from uservice_utils.logging import (
    configure_service_logging,
    ExtraLogger,
)
from uservice_utils.queue import (
    DefaultRetryPolicy,
    MessageActions,
    SimpleRabbitQueueWorker,
)

from core_result_checker import constants


class Worker(object):

    def __init__(self, swift_manager, queue_manager):
        get_logger(__name__).info("Service Started.")
        self.swift_manager = swift_manager
        self.queue_manager = queue_manager

    def __call__(self, message):
        logger = get_logger(__name__, message)
        logger.info("Got: %r", message)
        try:
            channel = message['channel']
            device = message['device']
            image_name = message['image_name']
            exit_code = message['exit_code']
        except KeyError as e:
            logger.error("Unable to unpack incoming message: %s", str(e))
            return MessageActions.Retry

        # check to see if adt-run reported infrastructure failures:
        if int(exit_code) in (16, 20, 100):
            logger.info(
                "Test run infrastructure failed (exit code %s), retrying.",
                exit_code
            )
            self.queue_manager.retry_test_run(message)
            return MessageActions.Acknowledge
        container_name = "core-{}-{}-{}".format(
            channel,
            device,
            image_name,
        ).replace('/', '_')
        tarball_name = container_name + "-results.tgz"

        # check to see if swift is consistent yet:
        if not self.swift_manager.container_contains_file(
            container_name,
            tarball_name
        ):
            logger.warning("Swift container not yet consistent.")
            self.queue_manager.retry_for_swift(message)
            return MessageActions.Acknowledge

        try:
            self.swift_manager.make_container_public(container_name)
        except RuntimeError as e:
            logger.error(
                "Unable to publish swift container '%s'. Error: %s.",
                container_name,
                e
            )
            return MessageActions.Retry
        logger.info("Swift container made public OK.")
        self.queue_manager.complete_test(message)
        return MessageActions.Acknowledge


def get_logger(name, extra={}):
    if logging.getLoggerClass() != ExtraLogger:
        logging.setLoggerClass(ExtraLogger)
    logger = logging.getLogger(name)
    e = constants.LOGGING_EXTRA.copy()
    e.update(extra)
    logger.set_extra_args(e)
    return logger


def read_config():
    parser = argparse.ArgumentParser(description='Core Result Checker.')
    parser.add_argument(
        '-c',
        '--conf',
        default='core-service.conf',
        help='Configuration file path'
    )
    args = parser.parse_args()

    # Load configuration options.
    config = configparser.ConfigParser()
    config.read(args.conf)
    return config


class SwiftManager(object):

    """A class that knows how to interact with swift."""

    def __init__(self, nova_config):
        self.config = nova_config

    def make_container_public(self, container_name):
        """Make a swift contianer public.

        Raises RuntimeError on failure, returns None otherwise.

        """
        swift = SwiftService(self.config)
        result = swift.post(
            container=container_name,
            options=dict(read_acl='.r:*')
        )
        if not result['success']:
            raise RuntimeError(
                "Unable to make swift container '%s' public: %s" % (
                    container_name,
                    result['error']
                )
            )

    def container_contains_file(self, container_name, file_name):
        swift = SwiftService(self.config)
        result = next(swift.stat(container_name, [file_name]))
        return result['success']


class QueueManager(object):

    """A class that knows how to retry a test payload."""

    def __init__(
        self, retry_test_fn, dead_letter_fn, input_fn, complete_fn,
        max_retries
    ):
        self.retry_test = retry_test_fn
        self.dead_letter = dead_letter_fn
        self.requeue_test = input_fn
        self.complete_test = complete_fn
        self.max_retries = max_retries
        self.swift_retry_seconds = 60

    def retry_test_run(self, payload):
        """Maybe retry a test payload. Due to adt-run reported failure.

        This will look at it's retry count, and either requeue it in the
        test queue, or insert it into the dead letter queue.

        """
        retry_count = int(payload.get('test_run_retry_count', '0'))
        if retry_count < self.max_retries:
            payload['test_run_retry_count'] = retry_count + 1
            self.retry_test(payload)
        else:
            logger = get_logger(__name__, payload)
            logger.error(
                "Test retry count exceeds maximum. Inserting into dead "
                "letter queue"
            )
            self.dead_letter(payload)

    def retry_for_swift(self, payload):
        """Maybe retry a test payload while waiting for swift.

        This will look at the first time the message was checked, and give a
        certain grace period. After that period is over, we will put the
        payload into the dead letter queue.

        """
        swift_check_key = 'swift_first_check_time'
        first_swift_check_time = payload.get(swift_check_key)
        if first_swift_check_time is None:
            first_swift_check_time = datetime.utcnow()
            payload[swift_check_key] = str(first_swift_check_time.timestamp())
        else:
            first_swift_check_time = datetime.fromtimestamp(
                float(first_swift_check_time)
            )
        now = datetime.utcnow()
        time_since_first_check = now - first_swift_check_time
        logger = get_logger(__name__, payload)
        if time_since_first_check.total_seconds() < self.swift_retry_seconds:
            logger.info(
                "Requeueing message since swift is not yet consistent"
            )
            self.requeue_test(payload)
        else:
            logger.error(
                "swift is still not consistent after %d seconds. Inserting "
                "message into dead letter queue.",
                self.swift_retry_seconds
            )
            self.dead_letter(payload)


def insert_payload_into_queue(connection, queue, payload):
    q = connection.SimpleQueue(queue)
    q.put(payload)
    q.close()


def insert_payload_into_fanout_exchange(connection, exchange_name, payload):
    with producers[connection].acquire(block=True) as producer:
        publisher = connection.ensure(
            producer,
            producer.publish,
            max_retries=3
        )
        exchange = kombu.Exchange(exchange_name, type="fanout")
        publisher(payload, exchange=exchange, declare=[exchange])


def main():
    config = read_config()
    log_path = os.path.abspath(
        os.path.join(__file__, '../../logs/core-result-checker.log'))
    configure_service_logging(
        log_path,
        config['logstash'] if 'logstash' in config else None
    )

    amqp_uris = config.get('amqp', 'uris').split()
    retry_policy = DefaultRetryPolicy(
        max_retries=3,
        dead_queue=constants.DEAD_LETTER_QUEUE
    )
    try:
        with kombu.Connection(amqp_uris) as connection:
            retry_payload = partial(
                insert_payload_into_queue,
                connection,
                constants.TEST_RETRY_QUEUE
            )
            deadletter_payload = partial(
                insert_payload_into_queue,
                connection,
                constants.DEAD_LETTER_QUEUE
            )
            requeue_payload = partial(
                insert_payload_into_queue,
                connection,
                constants.INPUT_QUEUE
            )
            complete_payload = partial(
                insert_payload_into_fanout_exchange,
                connection,
                constants.DONE_EXCHANGE,
            )
            queue_manager = QueueManager(
                retry_payload,
                deadletter_payload,
                requeue_payload,
                complete_payload,
                3
            )
            swift_manager = SwiftManager(config['nova'])
            worker = Worker(swift_manager, queue_manager)
            queue_monitor = SimpleRabbitQueueWorker(
                connection,
                constants.INPUT_QUEUE,
                worker,
                retry_policy,
            )
            queue_monitor.run()
    except KeyboardInterrupt:
        print("Bye")
